@@ -47,18 +47,19 @@ class GenericForeignFileWidget(Input):
             except MultipleObjectsReturned:
                 file_value = value
                 if bound_field and getattr(bound_field, 'form', None):
-                    value = bound_field.form.data.get("%s-0-id" % name)
+                    value = bound_field.form.data.get("%s-0-id" % bound_field.form.add_prefix(name))
             else:
                 file_value = value
                 value = obj.pk
         if rel_model and isinstance(value, (long, int)) or (isinstance(value, basestring) and value.isdigit()):
-            try:
-                obj = rel_model.objects.get(pk=value)
-            except ObjectDoesNotExist:
-                pass
-            else:
-                if getattr(obj, self.field.rel_file_field_name):
-                    file_value = getattr(obj, self.field.rel_file_field_name).name
+            if not obj or not file_value:
+                try:
+                    obj = rel_model.objects.get(pk=value)
+                except ObjectDoesNotExist:
+                    pass
+                else:
+                    if getattr(obj, self.field.rel_file_field_name):
+                        file_value = getattr(obj, self.field.rel_file_field_name).name
         else:
             obj = value
             try:
